/*
 * kgraft_patch_cve-2015-1421
 *
 * Fix for CVE-2015-1421, bsc#922004
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kallsyms.h>
#include <linux/printk.h>
#include <linux/bug.h>
#include <net/sctp/structs.h>

struct sctp_transport *(*kgr_sctp_assoc_lookup_paddr)(
		const struct sctp_association *asoc,
		const union sctp_addr *address);
void (*kgr_sctp_assoc_rm_peer)(struct sctp_association *asoc,
		struct sctp_transport *peer);
struct sctp_transport *(*kgr_sctp_assoc_add_peer)(
		struct sctp_association *asoc, const union sctp_addr *addr,
		const gfp_t gfp, const int peer_state);
void (*kgr_sctp_ulpq_flush)(struct sctp_ulpq *ulpq);
int (*kgr_sctp_auth_asoc_init_active_key)(struct sctp_association *asoc,
		gfp_t gfp);
struct sctp_tsnmap *(*kgr_sctp_tsnmap_init)(struct sctp_tsnmap *map, __u16 len,
		__u32 initial_tsn, gfp_t gfp);
void (*kgr_sctp_transport_reset)(struct sctp_transport *t);
void (*kgr_sctp_ssnmap_clear)(struct sctp_ssnmap *map);
int (*kgr_sctp_assoc_set_id)(struct sctp_association *asoc, gfp_t gfp);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "sctp_assoc_lookup_paddr", (void *)&kgr_sctp_assoc_lookup_paddr },
	{ "sctp_assoc_rm_peer", (void *)&kgr_sctp_assoc_rm_peer },
	{ "sctp_assoc_add_peer", (void *)&kgr_sctp_assoc_add_peer },
	{ "sctp_ulpq_flush", (void *)&kgr_sctp_ulpq_flush },
	{ "sctp_auth_asoc_init_active_key",
		(void *)&kgr_sctp_auth_asoc_init_active_key },
	{ "sctp_tsnmap_init", (void *)&kgr_sctp_tsnmap_init },
	{ "sctp_transport_reset", (void *)&kgr_sctp_transport_reset },
	{ "sctp_ssnmap_clear", (void *)&kgr_sctp_ssnmap_clear },
	{ "sctp_assoc_set_id", (void *)&kgr_sctp_assoc_set_id },
};

/* patched function */
void kgr_sctp_assoc_update(struct sctp_association *asoc,
			   struct sctp_association *new)
{
	struct sctp_transport *trans;
	struct list_head *pos, *temp;

	/* Copy in new parameters of peer. */
	asoc->c = new->c;
	asoc->peer.rwnd = new->peer.rwnd;
	asoc->peer.sack_needed = new->peer.sack_needed;
	asoc->peer.auth_capable = new->peer.auth_capable;
	asoc->peer.i = new->peer.i;
	kgr_sctp_tsnmap_init(&asoc->peer.tsn_map, SCTP_TSN_MAP_INITIAL,
			 asoc->peer.i.initial_tsn, GFP_ATOMIC);

	/* Remove any peer addresses not present in the new association. */
	list_for_each_safe(pos, temp, &asoc->peer.transport_addr_list) {
		trans = list_entry(pos, struct sctp_transport, transports);
		if (!kgr_sctp_assoc_lookup_paddr(new, &trans->ipaddr)) {
			kgr_sctp_assoc_rm_peer(asoc, trans);
			continue;
		}

		if (asoc->state >= SCTP_STATE_ESTABLISHED)
			kgr_sctp_transport_reset(trans);
	}

	/* If the case is A (association restart), use
	 * initial_tsn as next_tsn. If the case is B, use
	 * current next_tsn in case data sent to peer
	 * has been discarded and needs retransmission.
	 */
	if (asoc->state >= SCTP_STATE_ESTABLISHED) {
		asoc->next_tsn = new->next_tsn;
		asoc->ctsn_ack_point = new->ctsn_ack_point;
		asoc->adv_peer_ack_point = new->adv_peer_ack_point;

		/* Reinitialize SSN for both local streams
		 * and peer's streams.
		 */
		kgr_sctp_ssnmap_clear(asoc->ssnmap);

		/* Flush the ULP reassembly and ordered queue.
		 * Any data there will now be stale and will
		 * cause problems.
		 */
		kgr_sctp_ulpq_flush(&asoc->ulpq);

		/* reset the overall association error count so
		 * that the restarted association doesn't get torn
		 * down on the next retransmission timer.
		 */
		asoc->overall_error_count = 0;

	} else {
		/* Add any peer addresses from the new association. */
		list_for_each_entry(trans, &new->peer.transport_addr_list,
				transports) {
			if (!kgr_sctp_assoc_lookup_paddr(asoc, &trans->ipaddr))
				kgr_sctp_assoc_add_peer(asoc, &trans->ipaddr,
						    GFP_ATOMIC, trans->state);
		}

		asoc->ctsn_ack_point = asoc->next_tsn - 1;
		asoc->adv_peer_ack_point = asoc->ctsn_ack_point;
		if (!asoc->ssnmap) {
			/* Move the ssnmap. */
			asoc->ssnmap = new->ssnmap;
			new->ssnmap = NULL;
		}

		if (!asoc->assoc_id) {
			/* get a new association id since we don't have one
			 * yet.
			 */
			kgr_sctp_assoc_set_id(asoc, GFP_ATOMIC);
		}
	}

	/* SCTP-AUTH: Save the peer parameters from the new assocaitions
	 * and also move the association shared keys over
	 */
	kfree(asoc->peer.peer_random);
	asoc->peer.peer_random = new->peer.peer_random;
	new->peer.peer_random = NULL;

	kfree(asoc->peer.peer_chunks);
	asoc->peer.peer_chunks = new->peer.peer_chunks;
	new->peer.peer_chunks = NULL;

	kfree(asoc->peer.peer_hmacs);
	asoc->peer.peer_hmacs = new->peer.peer_hmacs;
	new->peer.peer_hmacs = NULL;

	/*
	 * FIX CVE-2015-1421
	 *
	 * -sctp_auth_key_put(asoc->asoc_shared_key);
	 */
	kgr_sctp_auth_asoc_init_active_key(asoc, GFP_ATOMIC);
}

static int kgr_patch_cve_2015_1421_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}
		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_cve_2015_1421_module_notify(struct notifier_block *nb,
		unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, "sctp"))
		return 0;

	ret = kgr_patch_cve_2015_1421_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_cve_2015_1421_module_nb = {
	.notifier_call = kgr_patch_cve_2015_1421_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_cve_2015_1421_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module("sctp")) {
		ret = kgr_patch_cve_2015_1421_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_cve_2015_1421_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_cve_2015_1421_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_cve_2015_1421_module_nb);
}
